****
*   TBWhile.prg
*   Browse database for specific key value
*   Copyright (c) 1990-1991 Nantucket Corp.  All rights reserved.
*
*   Compile:  clipper TBWhile /m/n/w
*   Link:     rtlink file TBWhile
*   Execute:  TBWhile
*

#include "Inkey.ch"
#include "Setcurs.ch"
#include "Error.ch"

// Separator strings for the browse display
#define MY_HEADSEP        ""
#define MY_COLSEP         "  "
#define MY_CHILD_HEADSEP   ""

***
*   TBWhile
*
PROCEDURE TBWhile()
	LOCAL bSaveHandler

	// Lazy man's error checking
	bSaveHandler := ERRORBLOCK({|x| BREAK(x)})

	BEGIN SEQUENCE
		USE Invoice INDEX Invoice NEW READONLY
		USE Detail  INDEX Detail  NEW READONLY

	RECOVER
		?? "Error opening file(s)"

		QUIT
	END

	// Restore the default error handler
	ERRORBLOCK(bSaveHandler)

	CLEAR SCREEN

	@ 1,0 SAY PADC( "Parent/Child Browse", ;
		MAXCOL() )
        FuzzScreen()
	PCBrowse(3, 6, 9, MAXCOL() - 6, ;
		SELECT("Invoice"), SELECT("Detail"), "ID")

	@ MAXROW(), 0
        CLEAR SCREEN

	RETURN

***
*   PCBrowse()
*   Create a TBrowse object and browse with it.
*
PROCEDURE PCBrowse( nTop, nLeft, nBottom, nRight, nParArea, nChildArea, cKeyField )
	LOCAL oBrowse // The TBrowse object
	LOCAL nKey // Keystroke
	LOCAL lMore // Loop control
	LOCAL nCursSave // Save state
	LOCAL xKeyVal
        LOCAL nLastRec := 0
        LOCAL lDisplayOnly := .T.

	// Make a TBrowse object for the parent workarea
	// that will fit inside box drawn below
	SELECT (nParArea)
	oBrowse := StockBrowseNew( nTop + 1, nLeft + 1, ;
		nBottom - 1, nRight - 1 )

	// Change the heading and column separators
	oBrowse:headSep := MY_HEADSEP
	oBrowse:colSep := MY_COLSEP

	// Draw a box around the browse
	@ nTop, nLeft CLEAR TO nBottom, nRight
	@ nTop, nLeft TO nBottom, nRight

	// Save cursor shape, turn the cursor off while browsing
	nCursSave := SETCURSOR(SC_NONE)

	// Main loop
	lMore := .T.
	DO WHILE (lMore)
		// Stabilize the browse
		ForceStable(oBrowse )
      
		// Banged against top or bottom of browse
		IF oBrowse:hitTop .OR. oBrowse:hitBottom
			TONE(125, 0)
		ENDIF
                // Make sure the Child display reflects detail items
                // specific to the current stock Number
                IF nLastRec # Invoice->( RECNO() )
                  lDisplayOnly := .T.
                  RefreshChild( cKeyField, nChildArea, nParArea, lDisplayOnly )
                ENDIF
                nLastRec := Invoice->( RECNO() )

		// Get a new keystroke
		nKey := INKEY(0)

		DO CASE
		CASE nKey == K_ESC
			// Esc means leave
			lMore := .F.

                CASE nKey == K_TAB
                        oBrowse:autolite := .F.
                        oBrowse:deHilite()
                        ForceStable( oBrowse )
                        oBrowse:autolite := .T.

                        lDisplayOnly := .F.
                        RefreshChild( cKeyField, nChildArea, nParArea, lDisplayOnly )

		OTHERWISE
			// Apply the key to the browse
			ApplyKey(oBrowse, nKey)
		ENDCASE
	ENDDO

	SETCURSOR(nCursSave)

	RETURN

***
*   BrowseWhile()
*   Browse database while the key field is matches xKey
*
*
PROCEDURE BrowseWhile( nTop, nLeft, nBottom, nRight, xKey, lDisplayOnly )
	LOCAL oBrowse // The TBrowse object
	LOCAL nKey // Keystroke
	LOCAL lMore // Loop control
	LOCAL nCursSave // Save state
        LOCAL aRect := {}
        LOCAL oCol
        LOCAL bCol
        LOCAL cHeader := PADR( "Item", 21 ) + PADR( 'Unit Cost', 10 ) + ;
                         PADR( "Qty", 6 )   + PADR( "Ext'd Amt", 10 )
        FIELD Item, UnitCost, Units IN Detail
	// Create another generic browse
        oBrowse := TBrowseDB( nTop + 1, nLeft + 1, ;
	                      nBottom - 1, nRight - 1 )

        // Now let's show'em a calculated Field
        bCol := { || PADR( Item, 20 ) + ' ' + ;
                   PADL( ALLTRIM( STR( UnitCost, 7, 2 ) ), 9 ) + ;
                   ' ' + PADL( ALLTRIM( STR( Units, 5, 0 ) ), 5 ) +;
                   ' ' + PADL( ALLTRIM( STR( UnitCost * Units, 7,2 ) ) ,9 ) }

        oCol := TBColumnNew( cHeader, bCol )
        oBrowse:addColumn( oCol )

	// Change the heading and column separators
        oBrowse:headSep := MY_CHILD_HEADSEP
        // oBrowse:colSep := MY_COLSEP

        // Change the movement methods
	oBrowse:goTopBlock := FirstKeyBlock(xKey)
        oBrowse:goBottomBlock := LastKeyBlock(xKey)
        oBrowse:skipBlock := KeySkipBlock(xKey)

        // Initialize AutoHi-Lite Instance Variable depending on
        // Invocation method
        oBrowse:autolite := !( lDisplayOnly )

        // Draw a box around the browse
        @ nTop, nLeft CLEAR TO nBottom, nRight
        @ nTop, nLeft TO nBottom, nRight

	// Save cursor shape, turn the cursor off while browsing
	nCursSave := SETCURSOR(SC_NONE)

	// Main loop
	lMore := .T.
	DO WHILE (lMore)

		// Stabilize the browse

		ForceStable(oBrowse)

		// Banged against top or bottom of browse
		IF oBrowse:hitTop .OR. oBrowse:hitBottom
			TONE(125, 0)
		ENDIF

		// Get a new keystroke
		nKey := INKEY(0)
                DO CASE
		  CASE nKey == K_ESC
			// Esc means leave
			lMore := .F.

                  CASE nKey == K_TAB
                        oBrowse:autolite := .F.
                        oBrowse:deHilite()
                        ForceStable( oBrowse )

                        lMore := .F.

                  OTHERWISE

			// Apply the key to the browse
			ApplyKey(oBrowse, nKey)

		ENDCASE
                
	ENDDO

        oBrowse:autolite := .F.
        oBrowse:deHilite()
        ForceStable( oBrowse )

	SETCURSOR(nCursSave)

	RETURN

***
*   FirstKeyBlock()
*   Create block that seeks to the first occurance
*   of the passed key.  Use scoping nature of blocks
*   to good use.
*
*
FUNCTION FirstKeyBlock( xKey )
	RETURN ({|| DBSEEK(xKey)})

***
*   LastKeyBlock()
*   Create block that seeks to the last occurance
*   of the passed key.  Use scoping nature of blocks
*   to good use.
*
*
FUNCTION LastKeyBlock( xKey )
	RETURN ({|| DBSEEK(NextNTXKey(xKey), .T.), DBSKIP(-1)})

***
*    NextNTXKey()
*    Return value of next key in database
*
*    WARNING! Does not work on logical values or
*    floating point numbers
*
*
FUNCTION NextNTXKey( xValue )
	LOCAL cType := VALTYPE(xValue)
	LOCAL xNext

	DO CASE
	CASE cType == "C"
		// Find next value by incrementing the ASCII value of
		// the last character in the string.
		//
		// NOTE: Assume last char is not 255 for simplicity
		xValue := STUFF(xValue, LEN(xValue), 1, ;
			CHR( ASC(RIGHT(xValue,1)) + 1 ))

	CASE cType == "N"
		// Since we can only handle integers, increment value
		// by one
		xValue++

	CASE cType == "D"
		// Increment date by one day
		xValue++

	OTHERWISE
		// Couldn't handle it

	ENDCASE

	RETURN (xNext)

****
*   KeySkipBlock()
*   Create skipBlock, putting scoping of blocks to good
*   use
*
FUNCTION KeySkipBlock( xKey )
	RETURN ({|nMove| KeySkipper(nMove, xKey)})





FUNCTION RefreshChild( cKeyField, nChildArea, nParArea, lDisplayOnly )



  	// Get current value of key field
  	LOCAL xKeyVal := FIELDGET(FIELDPOS(cKeyField))

  	SELECT (nChildArea)
  	SEEK xKeyVal

        IF lDisplayOnly
          KEYBOARD CHR( K_ESC )
        ENDIF

        BrowseWhile( MAXROW() / 2, 3, MAXROW()-1, MAXCOL()-3, xKeyVal, lDisplayOnly )

  	SELECT (nParArea)

RETURN (NIL)

****
*   KeySkipper()
*   Handle record movement requests from the TBrowse object
*   staying within one key.
*
FUNCTION KeySkipper( nMove, xKey )
	LOCAL nMoved

	nMoved := 0

	IF nMove == 0 .OR. LASTREC() == 0
		// Skip 0 (significant on a network)
		SKIP 0

	ELSEIF nMove > 0 .AND. RECNO() != LASTREC() + 1
		// Skip forward
		DO WHILE nMoved <= nMove .AND. ! EOF() .AND. ;
				&(INDEXKEY(0)) == xKey
			SKIP 1
			nMoved++
		ENDDO

		// All cases above lead to skipping one too many
		SKIP -1
		nMoved--

	ELSEIF nMove < 0
		// Skip backward
		DO WHILE nMoved >= nMove .AND. ! BOF() .AND. ;
				&(INDEXKEY(0)) == xKey
			SKIP -1
			nMoved--
		ENDDO

		// No phantom BOF() record, otherwise above loop
		// leads to skipping back one too far
		IF ! BOF()
			SKIP 1
		ENDIF

		nMoved++
	ENDIF

	RETURN (nMoved)

***
*   ForceStable()
*   Stabilize the browse
*
PROCEDURE ForceStable( oBrowse )

	DO WHILE .NOT. oBrowse:stabilize()
	ENDDO

	RETURN


***
*   StockBrowseNew()
*   Create a "stock" TBrowse object for the current workarea.
*
FUNCTION StockBrowseNew( nTop, nLeft, nBottom, nRight, nMode )
	LOCAL oBrowse
	LOCAL n, oCol

	// Start with a new oBrowse object from ToBrowseDB()
	oBrowse := TBrowseDB(nTop, nLeft, nBottom, nRight)

	// Add a column for each field in the current workarea
	FOR n := 1 TO FCOUNT()
		// Make a new column
    
		oCol := TBColumnNew( FIELDNAME(n), ;
			FIELDWBLOCK(FIELDNAME(n), SELECT()) )

		// Add the column to the oBrowse
		oBrowse:addColumn(oCol)

	NEXT

	RETURN (oBrowse)

***
*   ApplyKey()
*   Apply one keystroke to the browse.
*
*
PROCEDURE ApplyKey( oBrowse, nKey )

	DO CASE
	CASE nKey == K_DOWN
		oBrowse:down()

	CASE nKey == K_PGDN
		oBrowse:pageDown()

	CASE nKey == K_CTRL_PGDN
		oBrowse:goBottom()

	CASE nKey == K_UP
		oBrowse:up()

	CASE nKey == K_PGUP
		oBrowse:pageUp()

	CASE nKey == K_CTRL_PGUP
		oBrowse:goTop()

	CASE nKey == K_RIGHT
		oBrowse:RIGHT()

	CASE nKey == K_LEFT
		oBrowse:LEFT()

	CASE nKey == K_HOME
		oBrowse:home()

	CASE nKey == K_END
		oBrowse:end()

	CASE nKey == K_CTRL_LEFT
		oBrowse:panLeft()

	CASE nKey == K_CTRL_RIGHT
		oBrowse:panRight()

	CASE nKey == K_CTRL_HOME
		oBrowse:panHome()

	CASE nKey == K_CTRL_END
		oBrowse:panEnd()

	OTHERWISE
		// Let them know that they hit a bad key
		TONE(125, 0)

	ENDCASE

	RETURN

